import os
import random
import copy
import json


import numpy as np
from termcolor import colored
from glob import glob
import pickle


import torch
import torch.nn.functional as F
from torch.utils.data.dataset import Dataset



################################
# Demonstration
################################

def get_dataset(args, task_name, train):   
    train_data, test_data, new_test_data, train_action_gt, test_action_gt, new_test_action_gt, train_task_name, test_task_name, new_test_task_name, train_action_id, test_action_id, new_test_action_id, action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, max_goal_length, max_action_length, max_node_length, max_subgoal_length = gather_data(args, task_name)    
    train_dset = demo_dset(args, train_data, train_action_gt, train_task_name, train_action_id, action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, max_goal_length, max_action_length, max_node_length, max_subgoal_length)
    test_dset = demo_dset(args, test_data, test_action_gt, test_task_name, test_action_id, action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, max_goal_length, max_action_length, max_node_length, max_subgoal_length)
    new_test_dset = demo_dset(args, new_test_data, new_test_action_gt, new_test_task_name, new_test_action_id, action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, max_goal_length, max_action_length, max_node_length, max_subgoal_length)
    return train_dset, test_dset, new_test_dset


def collate_fn(data_list):
    graph_data = [data[0] for data in data_list]
    batch_goal_index = [data[1] for data in data_list]
    batch_valid_action_with_walk_index = [data[2] for data in data_list]

    if len(graph_data[0])==3:
        batch_graph_length = [d[0] for d in graph_data]
        batch_graph_input = [d[1] for d in graph_data]
        batch_file_name = [d[2] for d in graph_data]
    else:
        batch_graph_length = [d[0] for d in graph_data]
        batch_file_name = [d[1] for d in graph_data]

    if len(graph_data[0])==3:
        batch_demo_data = (
            np.arange(len(batch_graph_length)), 
            batch_graph_length, 
            batch_graph_input,
            batch_file_name
        )
    else:
        batch_demo_data = (
            np.arange(len(batch_graph_length)), 
            batch_graph_length,
            batch_file_name
        )

    return batch_demo_data, batch_goal_index, batch_valid_action_with_walk_index


def to_cuda_fn(data):
    batch_demo_data, batch_goal_index, batch_valid_action_with_walk_index = data

    if len(batch_demo_data)==4:
        batch_demo_index, batch_graph_length, batch_graph_input, batch_file_name = batch_demo_data

        batch_graph_input_class_objects = [[torch.tensor(j['class_objects']).cuda() for j in i] for i in batch_graph_input]
        batch_graph_input_object_coords = [[torch.tensor(j['object_coords']).cuda() for j in i] for i in batch_graph_input]
        batch_graph_input_states_objects = [[torch.tensor(j['states_objects']).cuda() for j in i] for i in batch_graph_input]
        batch_graph_input_mask_object = [[torch.tensor(j['mask_object']).cuda() for j in i] for i in batch_graph_input]
        
        batch_graph_input = { 'class_objects': batch_graph_input_class_objects, 
                                'object_coords': batch_graph_input_object_coords,
                                'states_objects': batch_graph_input_states_objects,
                                'mask_object': batch_graph_input_mask_object}

    else:
        batch_demo_index, batch_graph_length, batch_file_name = batch_demo_data
    

    batch_goal_index = [torch.tensor(i).cuda().long() for i in batch_goal_index]
    batch_valid_action_with_walk_index = [torch.tensor(i).cuda().long() for i in batch_valid_action_with_walk_index]
    
    if len(batch_demo_data)==4:
        batch_demo_data = (
            batch_demo_index,
            batch_graph_length, 
            batch_graph_input,
            batch_file_name
        )
    else:
        batch_demo_data = (
            batch_demo_index,
            batch_graph_length,
            batch_file_name
        )

    return batch_demo_data, batch_goal_index, batch_valid_action_with_walk_index

def one_hot(states, graph_node_states):
    one_hot = np.zeros(len(graph_node_states))
    for state in states:
        one_hot[graph_node_states[state]] = 1
    return one_hot

def gather_data(args, task):    
    meta_data_path = 'dataset/watch_data/metadata.json'
    data_path_new_test = 'dataset/watch_data/action/new_test_task_' + task + '_strategy.json'
    data_path_test = 'dataset/watch_data/action/test_task_' + task + '_strategy.json'
    data_path_train = 'dataset/watch_data/action/train_task_' + task + '_strategy.json'

    
    with open(data_path_new_test, 'r') as f:
        new_test_data = json.load(f)
        
    with open(data_path_test, 'r') as f:
        test_data = json.load(f)
    
    #if os.path.exists(data_path):
        #print('load gather_data, this may take a while...', data_path)
    with open(data_path_train, 'r') as f:
        data = json.load(f)
      
    # temporarily set data_path to test data to test training
    train_data = data

    with open(meta_data_path, 'r') as f:
        data = json.load(f)
            
        action_predicates = data['action_predicates']
        all_action = data['all_action']
        all_object = data['all_object']
        
        goal_objects = data['goal_objects']
        goal_targets = data['goal_targets']
        goal_predicates = data['goal_predicates']
            
        graph_class_names = data['graph_class_names']
        graph_node_states = data['graph_node_states']

        max_goal_length = data['max_goal_length']
        max_action_length = data['max_action_length']
        max_node_length = data['max_node_length']
            

    ## -----------------------------------------------------------------------------
    ## add action, goal, and graph node index
    ## -----------------------------------------------------------------------------
    max_subgoal_length = 1
    
    train_task_name = np.array(train_data['task_name'])
    
    test_task_name = np.array(test_data['task_name'])
    
    new_test_task_name = np.array(new_test_data['task_name'])
    
    
    for traintest in [train_data, test_data, new_test_data]:
        for data in traintest['goal']:
            ## goal
            goal_index = []
            subgoal_dict = {}
            for subgoal in data:
                goal_index.append(goal_predicates[subgoal])

                if goal_predicates[subgoal] not in subgoal_dict:
                    subgoal_dict[goal_predicates[subgoal]] = 1
                else:
                    subgoal_dict[goal_predicates[subgoal]] += 1

            this_max_subgoal_length = np.max(list(subgoal_dict.values()))
            if this_max_subgoal_length>max_subgoal_length:
                max_subgoal_length = this_max_subgoal_length
            

            goal_index.sort()
            if len(goal_index) < max_goal_length:
                for i in range(max_goal_length-len(goal_index)):
                    goal_index.append(0)                  

        ## action gt       
        for i in range(len(traintest['action_gt'])): # len(traintest['action_gt'])
            action_name = traintest['action_gt'][i][0].split(' ')[0]
            object_name = traintest['action_gt'][i][0].split(' ')[1]
            predicate_name = ' '.join([action_name, object_name])
            traintest['action_gt'][i] = action_predicates[predicate_name]
                
        ## action        
        valid_action_with_walk_index = []
        for i in range(len(traintest['valid_action_with_walks'])):   
            actions_index = []            
            for actions in traintest['valid_action_with_walks'][i]:
                if actions!='None':
                    action_name = actions[0].split(' ')[0]
                    object_name = actions[0].split(' ')[1]
                    predicate_name = ' '.join([action_name, object_name])
                else:
                    predicate_name = actions
                actions_index.append(action_predicates[predicate_name])
                
            traintest['valid_action_with_walks'][i] = actions_index
                      
    print(len(train_data['action_gt']),np.array(train_data['action_gt']), type(train_data['action_gt']))
    
    
    train_action_gt =  np.array(train_data['action_gt'])    
    test_action_gt = np.array(test_data['action_gt'])    
    new_test_action_gt = np.array(new_test_data['action_gt'])
    
    train_action_id =  np.array(train_data['action_id'])    
    test_action_id = np.array(test_data['action_id'])    
    new_test_action_id = np.array(new_test_data['action_id'])
    
    train_data = np.array(train_data['valid_action_with_walks'])        
    test_data = np.array(test_data['valid_action_with_walks'])        
    new_test_data = np.array(new_test_data['valid_action_with_walks'])
    
    
    print('--------------------------------------------------------------------------------')
    print('train_data', len(train_data), train_data.shape)
    print('test_data', len(test_data), train_data.shape)
    print('new_test_data', len(new_test_data), train_data.shape)
    print('--------------------------------------------------------------------------------')
    print('train_gt', len(train_action_gt), train_action_gt.shape)
    print('test_gt', len(test_action_gt), test_action_gt.shape)
    print('new_test_gt', len(new_test_action_gt), new_test_action_gt.shape)
    print('--------------------------------------------------------------------------------')
    print('train_task_name', len(train_task_name), train_task_name.shape)
    print('test_task_name', len(test_task_name), test_task_name.shape)
    print('new_test_task_name', len(new_test_task_name), new_test_task_name.shape)
    print('--------------------------------------------------------------------------------')
        
    return train_data, test_data, new_test_data, train_action_gt, test_action_gt, new_test_action_gt, train_task_name, test_task_name, new_test_task_name, train_action_id, test_action_id, new_test_action_id, action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, max_goal_length, max_action_length, max_node_length, max_subgoal_length

class demo_dset(Dataset):

    def __init__(
            self,
            args,
            data, 
            gt,
            task_name,
            action_id,
            #action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, graph_class_names, graph_node_states, max_goal_length, max_action_length, max_node_length, max_subgoal_length):
            action_predicates, all_action, all_object, goal_objects, goal_targets, goal_predicates, max_goal_length, max_action_length, max_node_length, max_subgoal_length):


        self.inputtype = args.inputtype
        self.multi_classifier = args.multi_classifier
        self.data = data
        self.gt = gt
        self.task_name = task_name
        self.action_id = action_id
        
        self.max_action_len = 82
        self.action_predicates = action_predicates
        self.all_action = all_action
        self.all_object = all_object
        
        self.goal_objects = goal_objects
        self.goal_targets = goal_targets
        self.goal_predicates = goal_predicates
        self.num_goal_predicates = len(goal_predicates)

        self.max_goal_length = max_goal_length
        self.max_action_length = max_action_length
        self.max_subgoal_length = max_subgoal_length

        if self.inputtype=='graphinput':
            self.graph_class_names = graph_class_names
            self.graph_node_states = graph_node_states
            self.num_node_states = len(graph_node_states)
            self.max_node_length = max_node_length


        print('-----------------------------------------------------------------------------')
        print('num_goal_predicates', self.num_goal_predicates)
        print('max_goal_length', self.max_goal_length)
        print('max_action_length', max_action_length)
        
        if self.inputtype=='graphinput':
            print('num_node_states', self.num_node_states)
            print('max_node_length', max_node_length)
        print('-----------------------------------------------------------------------------')

        
    def __getitem__(self, index):
        data = self.data[index]
        gt = self.gt[index]
        task_name = self.task_name[index]
        action_id = self.action_id[index]
        
        return data, gt, task_name, action_id

    def __len__(self):
        return len(self.data)

    def _preprocess_one_data(self, data):       
        action_gt = data['action_gt']
        valid_action_with_walk_index = data['valid_action_with_walks']
        action_length = len(valid_action_with_walk_index)
        inputdata = (action_length, 'actions')
        data = [inputdata, action_gt, valid_action_with_walk_index]
        return data
    
        
